#!c:\perl\bin\perl

# ========================================================================
# Hello! This is the source code for the ELZA interpreter.
# Make sure you edit elza.def before using this script.
# This script is placed in the public domain. Use as you see fit.
# The homepage of the project is http://phiphi.hypermart.net .
# The contact email is philip_stoev@iname.com. Feedback is ALWAYS welcome.
# ========================================================================

# This is required for TCP/IP functionality
use Socket;

# This is required for parsing command line options
use Getopt::Std;

# Those are required for interprocess communication
use IPC::Open2;

$DEF{version} = '1.4.5';	# ELZA Version

# Analyse command line options and place them in the cmdopt substs.
getopts("a:b:c:d:e:f:g:h:i:j:k:l:m:n:o:p:q:r:s:t:u:v:w:x:y:z:I:S:", \%cmdopts);
foreach $opt (%cmdopts) {$VAR{'cmdopt'.$opt} = $cmdopts{$opt};}

# If we are forked, -i and -I will contain the sequence
# number of the process and the total number of processes, respectively.
$VAR{instance} = $cmdopts{i};
$VAR{instances} = $cmdopts{I};

# Read various default values from the defaults file
if ($cmdopts{d} eq '') {$VAR{def_file} = 'elza.def';} else {$VAR{def_file} = $cmdopts{d};}
open (DEFAULTS, "<$VAR{def_file}") or shout ('critical', "Unable to open defaults file $VAR{def_file}: $!");
shout ('debug', "INIT: Now reading defaults from $VAR{def_file}.");
foreach $Line (<DEFAULTS>) {eval $Line;}
close DEFAULTS;
if ($DefaultsReviewed != 1) {shout ('critical', "Please review the defaults in $DEFAULTS before running the ELZA.");}
# This is not very fast, but since we do it only once, we leave it like that

# Adjust shouting level to the one specified on the command line
if ($cmdopts{S} ne '') {$VAR{shout} = $cmdopts{S};}
# Initally, no successful requests
$VAR{tot_req} = 0;		$SuccessfulRequests = 0;
$VAR{tot_sent} = 0;			$VAR{tot_rcv} = 0; 
$VAR{tot_ok} = 0;		$RedirRequests = 0;
$VAR{tot_pass} = 0;
$SocketOpened = 0;

# Unbuffer STDOUT
$| = 1;

$LastPosition = 0;

# If we are called via GET / HEAD ...
if (($ENV{REQUEST_METHOD} ne '') && ($ENV{REQUEST_METHOD} ne 'POST')) {
	print STDOUT "Content-type: text/plain\n\n";
	print STDOUT "Please execute ELZA with POST.\n";
	exit;
}

# If we are called via a POST ...
if ($ENV{REQUEST_METHOD} eq 'POST') {
	# Read the fields of the form
	read(STDIN, $buffer, $ENV{CONTENT_LENGTH});
	@pairs = split(/&/, $buffer);
	foreach $pair (@pairs) {
		($name, $value) = split(/=/, $pair);
		$value =~ tr/+/ /;
		$value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
		$FORM{$name} = $value;
	}
	if ($FORM{shout} eq 'silent') {
		$VAR{shout} = 'silent';
	} else {
		shout ('raw', 'Content-type: text/plain\n\n');
		shout ('normal', "--- Hello. This is the ELZA interpreter version $DEF{version}.");
		shout ('normal', "--- Running in remote mode. Called from $ENV{REMOTE_ADDR}. Shout: $FORM{shout}.");
	}
	# The file to execute is within the 'elza-web-script' field
	$SCRIPT = $FORM{'elza-web-script'};

} else {
	# The file to execute is the first argument of the command line
	shout ('info', "Hello. This is the ELZA interpreter version $DEF{version}.");
	if ($VAR{instance} ne '') {
		shout ('info', "I am instance $VAR{instance} of total $VAR{instances} instances.");
	}

	$SCRIPT = $cmdopts{s};

	if ($SCRIPT eq '') {
		print STDOUT "??? Please enter script to execute: ";
		$SCRIPT = <STDIN>;
		$SCRIPT =~ s/[\r\n]//g;
	}

	if ($SCRIPT eq '') {
		shout ('critical',"No script name was entered.");
	}
}

$startclock = time();

open (SCRIPT,"<$SCRIPT") or shout ('critical', "Unable to open script $SCRIPT: $!");
shout ('info',"Now processing commands from $SCRIPT");

$IsWebEnabled = 0;
$ProcName = '';
# Process the main command file, labels, procs, etc.
$LineNum = 0;
while (<SCRIPT>) {
	# Strip CR/LF, whitespace, comments
	s/[\r\n]//g;	s/^[ \t]*//;
	next if(/^\#/);	next if( $_ eq '');
	if ($_ =~ /^label (.*)/) {
		if ($ProcName ne '') {
			shout ('critical',"You can not define a label inside a procedure.");
		}

		if ($LABEL{$1} ne '') {
			shout ('critical',"You can not define a label more than once.");
		} else {
			shout ('debug', "PARSE: Label $1");
			$LABEL{$1} = $LineNum - 1;
		}
		next;
	}

	if ($_ =~ /^proc (.*)/) {
		if ($ProcName ne '') {
			shout ('critical', "Nesting of procedure definitions is not allowed.");
		}
		shout ('debug', "PARSE: Proc start $1 .");
		$ProcName = $1;
		@PROC{$ProcName} = ();
		next;
	}

	if ($_ =~ /^endproc (.*)/) {
		if ($1 ne $ProcName) {
			shout ('critical', "Statement \'$_\' does not match the last 'proc' directive.");
		}
		shout ('debug', "PARSE: Proc end $1 .");
		$ProcName = '';
		next;
	}

	if ($_ eq 'elza-web-enabled') {
		$IsWebEnabled = 1;
		shout ('debug', "PARSE: Found elza-web-enabled tag.");
		next;
	}

	if ($ProcName ne '') {
		push @{$PROC{$ProcName}}, $_;
		shout ('debug', "PARSE: Pushing \' $_ \' into proc $ProcName");
	} else {
		push @RawScript, $_;
		shout ('debug', "PARSE: Pushing \' $_ \' into script body");
		$LineNum ++;
	}
}

if ($ProcName ne '') {shout ('critical', "An 'endproc' statement is missing.");}

# End processing main command file

if (($ENV{REQUEST_METHOD} eq 'POST') && ($IsWebEnabled == 0)) {
	# The script was called via POST, but the command file was not
	# specifically marked. This is to avoid attempts to execute
	# files like /etc/passwd, etc.
	print "Content-type: text/plain\n\n";
	print "!!! This file does not have an elza-web-enabled tag.\n";
	exit;
}

# Begin processing commands
$ScriptEnd = $#RawScript;
$LineNum = 0;
while  ($LineNum <= $ScriptEnd) {
	# Process the command
	&ProcessCommand ($RawScript[$LineNum]);
	$LineNum ++;
}
close ExtReader;
close ExtWriter;
# End command processing loop

sub ProcessCommand {

	my @TempData;
	$TheCommand = $_[0];

	shout ('debug', "PROCESS: Now working on \' $TheCommand \'");

	# Split the command, placing parts in Param and ParamAfter

	my @Param = split(' ',$TheCommand);

	if (($Param[1] ne 'beforerequest') && ($Param[0] ne 'foreach')){
	for $PrmCnt (1..$#Param) {
		# Perform substitution of FORM values
		if ($ENV{REQUEST_METHOD} eq 'POST') {
			while ($Param[$PrmCnt] =~ /\$\$(.*?)\$\$/g) {
				$FormSubst = $1;
				next if (!(exists $FORM{$1}));
				$Param[$PrmCnt] =~ s/\$\$$FormSubst\$\$/$FORM{$FormSubst}/g;
				shout ('debug', "PROCESS: Form subst \' $key \' with \' $value \'.");
			}
		}

		# Perform dynamic SUBST substitution
		while ($Param[$PrmCnt] =~ /\@\@(.*?)\@\@/g) {
			$DynSub = $1;
			if ($SUBST{$DynSub} > $#{$SUBARRAY{$DynSub}}) {$SUBST{$DynSub} = 0;}
			$ArrayValue = ${$SUBARRAY{$DynSub}}[$SUBST{$DynSub}];
			if ($Param[$PrmCnt] =~ s/\@\@$DynSub\@\@/$ArrayValue/g) {
				shout ('debug', "PROCESS: Dynamic subst \' $key \' with \' $ArrayValue \'.");
			}
			if ($Param[0] ne 'print') {$SUBST{$DynSub}++;} 
		}
		
		# Static substs
		if ($Param[$PrmCnt] =~ s/\$\$(.*?)\$\$/$SUBST{$1}/g) {
			shout ('debug', "PARSE: Some substs (last was $1) were substituted.");
		}
		# Variables
		$Param[$PrmCnt] =~ s/\%\%(.*?)\%\%/$VAR{$1}/g;
	}
	}

	# @ParamAfter holds the entire line after the N-th parameter inclusive.
	my @ParamAfter = ();
	for $PrmCnt (0..$#Param) {
		$ParamAfter[$PrmCnt] = join(' ',@Param[$PrmCnt..$#Param]);
	}

	# Begin syntax parsing
	if ($Param[0] eq 'split') {
		$PartCount = 2;
		@Array = split ($Param[2],$SUBST{$Param[1]});
		foreach $Part (@Array) {
			$PartCount++;
			next if ($Param[$PartCount] eq '');
			$SUBST{$Param[$PartCount]} = $Part;
		}
		return;
	} elsif ($Param[0] eq 'printvars') {

		foreach $mojo (keys %VAR) {
			print "VAR: $mojo - $VAR{$mojo}\n";
		}

	} elsif ($Param[0] eq '2waywrite') {
		print ExtWriter "$ParamAfter[1]\n";
		return;
	} elsif (($Param[0] eq 'subst') || ($Param[0] eq 'field') || 
		($Param[0] eq 'var') || ($Param[0] eq 'cookie') ||
		($Param[0] eq 'header')) {

		$TempValue = '';
		$ToBeDeleted = 0;

		if ($Param[2] eq '=') {
			$TempValue = $ParamAfter[3];
		} elsif ($Param[2] eq 'f=') {
			$TempValue = $FLD{$ParamAfter[3]};
		} elsif ($Param[2] eq 'b=') {
			$TempValue = encode_base64($ParamAfter[3]);
		} elsif ($Param[2] eq 'c=') {
			$TempValue = $COOK{$ParamAfter[3]};
		} elsif ($Param[2] eq 'refresh') {
			# We will now try to find a "refresh" tag.
			$TagCode = 'refresh';$TagMethod = 'number';$TagMatch= '1';
			&FindTag;
			if ($FoundTag eq '%%NONE%%') {
				shout ('error', "Did not find an URL to refresh to. ");
			} else {
				($Garbage, $TempValue) = split (/; url=/i,$FoundTag);
			}
		} elsif ($Param[2] eq '-') {
			$ToBeDeleted = 1;
			shout ('debug', "OPER: Deleting $Param[0] \'$Param[1]\' .");
		} elsif ($Param[2] eq 'between') {
			$BeginHTML = $Param[3];
			$EndHTML = $Param[4];
			&FindHTML;
			if ($FoundHTML eq '%%NONE%%') {
				shout ('error', "Unable to locate HTML between $Param[3] and $Param[4]. ");
				return;
			} else {
				$TempValue = $FoundHTML;
				shout ('verbose', "HTML for $Param[1] located. ");
			}
		} elsif ($Param[2] eq 'line') {
			$TempValue = &FindLine($Param[3]);
			return if (!(defined $TempValue));
		} elsif ($Param[2] eq 'strip') {
			$TempValue = $ParamAfter[3];
			$TempValue =~ s/<(.*?)>//g;
			$TempValue =~ s/[\(\)]//g;
			$TempValue =~ s/[\r\n]//g;
			$TempValue =~ s/^[ \t]*//;
			$TempValue =~ s/[ \t]*$//;
		} elsif ($Param[2] eq '2wayread') {
			$TempValue = <ExtReader>;
			$TempValue =~ s/[\r\n]//g;
		} elsif ($Param[2] eq '@') {
			@TempArray = ();
			for $Element (3..$#Param) {
				push @TempArray, $Param[$Element];
			}
		} elsif ($Param[2] eq '%') {
			@TempArray = ();
			# Push the data from the file specified
			$VarFile = $Param[3];
			open (DATAFILE, "<$VarFile") or shout ('critical', "Unable to open data file $VarFile: $! .");
			while (<DATAFILE>) {
				$_ =~ s/[\r\n]//g;
				next if ($_ eq '');
				push @TempArray, $_;
			}
			close DATAFILE;
		} elsif ($Param[2] eq 'split') {
			@TempArray = split($Param[3], $ParamAfter[4]);
		} elsif ($Param[2] eq 'random') {
			$Range = $Param[4] - $Param[3];
			srand;
			$TempValue = ($Param[3] +  rand $Range);
			$TempValue = sprintf ("%d", $TempValue);
		} elsif ($Param[2] eq 'from') {
			$TagCode = $Param[5];
			$TagMethod = $Param[6];
			$TagMatch = $ParamAfter[7];
			&FindTag;
			if ($FoundTag eq '%%NONE%%') {
				shout ('error', "Unable to locate value for $Param[1] in $ParamAfter[1] .");
				return;
			} else {
				$TempValue = $TAGA{$Param[3]};
				shout ('verbose', "--- Tag for $Param[1] located - $TempValue.");
			}
		} elsif ($Param[2] eq 'raw') {
			$DATAFILE = "<$Param[3]";
			open DATAFILE or shout ('critical', "Unable to open data file $Param[3]: $!");
			binmode DATAFILE;
			read (DATAFILE, $TempValue, 1000000);
			close DATAFILE;
		} elsif ($Param[2] eq '?') {
			if ($ENV{REQUEST_METHOD} eq 'POST') {
				# If we are called via POST, we will not get value
				# from STDIN, but from %FORM
				$TempValue = $FORM{$Param[1]};
			} else {
				print STDOUT "??? Please enter value for $Param[0] \'$Param[1]\' [$ParamAfter[3]]: ";
				$TempValue = <STDIN>;
				$TempValue =~ s/[\r\n]//g;
				if ($TempValue eq '') { $TempValue = $ParamAfter[3]};
			}
		} elsif ($Param[2] eq '>') {
			shout ('verbose', "Overflowing $Param[0] \'$Param[1]\' with $Param[3] characters.");
			$TempValue = 'A' x $Param[3];
		} elsif ($Param[2] eq '$') {
			if ($Param[1] ne '%ALL%') {
				$TagCode = 'hidden';
				$TagMethod = 'name';
				$TagMatch = $Param[1];
				&FindTag;
				if ($FoundTag ne '%%NONE%%') {
					$FLD{$Param[1]} = $FoundTag;
				} else {
					shout ('error', "Hidden form field $Param[1] not found.");
				}
			} else {
				$TagEnumerate = 1;
				$TagCode = 'hidden';
				&FindTag;
				$TagEnumerate = 0;
			}
			return;
		}

		if ($Param[0] eq 'subst') {
			if ($ToBeDeleted > 0) {
				delete $SUBST{$Param[1]};
				delete $SUBTYPE{$Param[1]};
			} elsif (($Param[2] eq '@') || ($Param[2] eq '%') || ($Param[2] eq 'split')) {
				$SUBTYPE{$Param[1]} = 'array';
				$SUBST{$Param[1]} = 0;
				@{$SUBARRAY{$Param[1]}} = @TempArray;
			} else {
				$SUBST{$Param[1]} = $TempValue;
			}
			return;	
		} elsif ($Param[0] eq 'field') {
			if ($Param[1] eq '%BOGUS%') {
				$BogusField = 1;
				$BogusFieldSize = $Param[3];
				$BogusTextSize = $Param[4];
			} elsif ($ToBeDeleted > 0) {
				delete $FLD{$Param[1]};
			} else {
				$FLD{$Param[1]} = $TempValue;
			}
			$AnyFields = 1;
			return;
		} elsif ($Param[0] eq 'header') {
			if ($ToBeDeleted > 0) {
				$HEADER{$Param[1]} = '';
			} else {
				$HEADER{$Param[1]} = $TempValue;
			}
			return;
		} elsif ($Param[0] eq 'var') {
			$VAR{$Param[1]} = $TempValue;
			if ($Param[1] eq "rawfile") {
				unlink $VAR{rawfile};
			} elsif ($Param[1] eq "proxy") {
				shout ('verbose', "Bouncing through $Param[3] on port $Param[4]. ");
				$VAR{proxyhost} = $Param[3];
				$VAR{proxyport} = $Param[4];
				&CloseTheSocketForMe;
			} elsif (($Param[1] eq "proxyhost") || ($Param[1] eq "proxyport")) {
				&CloseTheSocketForMe;
			} elsif ($Param[1] eq "sockshost") {
				&CloseTheSocketForMe;
			} elsif ($Param[1] eq 'dictionary'){
				$DICTFILE = "<$TempValue";
				open DICTFILE or shout ('critical', "Unable to open dictionary $TempValue: $! .");
			} elsif ($Param[1] eq 'autopass') {
				$VAR{password} = $VAR{autopass};
			} elsif ($Param[1] eq 'password') {
				$VAR{autopass} = '';
			} elsif ($Param[1] eq 'autoresume') {
				# If we are about to resume a previous attack,
				# retrieve the current status
				$TEMPFILE = $VAR{tmpfile}.$VAR{instance};
				open TEMPFILE;
				$LastPosition = <TEMPFILE>;
				if ($LastPosition eq '') {
					$LastPosition = 1;
				} else {
					$LastPosition++;
					shout ('normal', "--- We will start from data line number $LastPosition .");
				}
				close TEMPFILE;
			} elsif ($Param[1] eq '2wayproc') {
				$ExtPid = open2(*ExtReader, *ExtWriter, $TempValue );
				shout ('debug',"OPEN2: Two-way process with PID $pid spawned.");
				select(ExtWriter); $| = 1; select(STDOUT);
			}
			return;
		} elsif ($Param[0] eq 'cookie') {
			if ($ToBeDeleted > 0) {
				if ($Param[1] eq '%ALL%') {
					undef %COOK;
				} else {
					delete $COOK{$Param[1]};
				}
			} elsif ($Param[2] eq 'path') {
				$COOKP{$Param[1]} = $TempValue;
			} elsif ($Param[2] eq 'domain') {
				$COOKD{$Param[1]} = $TempValue;
			} else {
				$COOK{$Param[1]} = $TempValue;
			}
			return;
		} else {
			shout ('critical', "Syntax errror: $ParamAfter[0] .");
		}
	} elsif (($Param[0] eq 'get') || ($Param[0] eq 'post') || ($Param[0] eq 'head') || ($Param[0] eq 'request')) {
		if ($Param[0] ne 'request') {
			$VAR{method} = uc($Param[0]);
		} else {
			$VAR{method} = uc($VAR{method});
		}

		if ($Param[1] eq "url") {
			# Requesting an ordinary URL. Very simple.
			$VAR{url} = $ParamAfter[2];

		} elsif ($Param[1] eq "refresh" ) {
			# We will now try to find a "refresh" tag.
			$TagCode = 'refresh';
			$TagMethod = 'number';
			$TagMatch= '1';
			&FindTag;
			if ($FoundTag eq '%%NONE%%') {
				shout ('error', "Did not find an URL to refresh to. ");
			} else {
				$VAR{url} = $FoundTag;
				shout ('verbose', "<<< Refresh to $VAR{url}");
			}
		} else {
			# Requesting and URL contained within some tag

			# Look for a <base> tag.
			shout ('debug', "HTML: Looking for a <BASE> tag.");
			$TagCode = 'base';
			$TagMethod = 'number';
			$TagMatch= '1';
			&FindTag;
			if ($FoundTag ne '%%NONE%%') {
				shout ('verbose', "Accepted base URL $VAR{base} .");
				$VAR{base} = $FoundTag;
			} else {
				shout ('debug', "HTML: No <BASE> tag found.");
				$VAR{base} = '';
			}
		
			# Look for the tag specified in the command
			$TagCode = $Param[1];
			$TagMethod = $Param[2];
			$TagMatch = $ParamAfter[3];
			&FindTag;
			if ($FoundTag eq '%%NONE%%') {
				shout ('error', "Unable to locate tag: $ParamAfter[1] .");
				return;
			} else {
				shout ('verbose', "--- URL for $Param[0] located: $FoundTag .");
			}

			# If the next URL is not absolute, add <base>
			if ($FoundTag =~ /($DEF{http}|$DEF{https})/) {
				$VAR{url} = $FoundTag;
			} else {
				$VAR{url} = $VAR{base}.$FoundTag;
			}
		}
	
		# Send the next request.
		&SendRequestToHost;
		%FLD = ();
		shout ('debug', "REQUEST: We are done with \' $TheCommand \'");
		return;
	} elsif ($Param[0] eq 'pause') {
		print STDOUT "--- Ran into pause command. Hit ENTER to continue ...";
		$Bogus = <STDIN>;
		return;
	} elsif ($Param[0] eq "sleep") {
		shout ('verbose', "Ran into sleep command. Sleeping for $Param[1] second(s) ...");
		sleep $Param[1];
		return;
	} elsif ($Param[0] eq "exec") {
		shout ('verbose', "Executing $ParamAfter[1] ...");
		eval($ParamAfter[1]);
		return;
	} elsif ($Param[0] eq 'if') {
		if ( MakeComparison ($Param[1], $Param[2], $Param[3]) == 1) {
			&ProcessCommand ($ParamAfter[4]);
		}
		return;
	} elsif ($Param[0] eq 'goto') {
		if ($Param[2] eq 'if') {
			if ( MakeComparison ($Param[3], $Param[4], $ParamAfter[5]) == 1) {
				$LineNum = $LABEL{$Param[1]};
				shout ('debug', "GOTO: Jumping to line with internal number $LineNum (with if condition)");
			}
		} else {
			$LineNum = $LABEL{$Param[1]};
			shout ('debug', "GOTO: Jumping to line with internal number $LineNum");
		}
		return;
	# End processing goto directive, begin processing call directive
	} elsif ($Param[0] eq 'call') {
		# Begin processing 'call if' directives
		if ($Param[2] eq 'if') {
			if (MakeComparison ($Param[3],$Param[4],$ParamAfter[5]) == 1) {
				&ExecProc ($Param[1]);
			} else {
				shout ('debug', "CALL: $Param[1] not executed because condition is not true.");
			}
			return;
		} elsif ($Param[2] eq '') {
			&ExecProc ($Param[1]);
			return;
		} elsif ($#{$PROC{$Param[1]}} == -1) {
			shout ('critical', "Proc $Param[1] not defined.");
		} elsif ($Param[3] eq '@') {
			# Push the array specified in the directive
			for $Element (4..$#Param) {
				$SUBST{$Param[2]} = $Param[$Element];
				&ExecProc ($Param[1]);
			}
		} elsif ($Param[3] eq 'eachfield') {
			# Rotate for every field available
			foreach $Value (keys %FLD) {
				$SUBST{$Param[2]} = $Value;
				&ExecProc ($Param[1]);
			}
			return;
		} elsif ($Param[3] eq '#') {
			for $Value ($Param[4]..$Param[5]) {
				$SUBST{$Param[2]} = $Value;
				&ExecProc ($Param[1]);
			}
			return;
		} elsif ($Param[3] eq '?') {
			# Read from STDIN
			$ValueCount = 1;
			print STDOUT "??? Please enter values for \'$Param[2]\', then blank line to continue:\n";
			print STDOUT "??? Value number $ValueCount: ";
			RSTD: while (<STDIN>) {
				s/[\r\n]//g;
				if ($_ eq '') {	last RSTD;}
				$SUBST{$Param[2]} = $_;
				&ExecProc ($Param[1]);
				$ValueCount++;
				print STDOUT "??? Value number $ValueCount: ";
			}
			if ($ValueCount == 1) {
				shout ('error', "No values entered. Skipping procedure \' $Param[1] \'.");
			}
			return;

		} elsif (($Param[3] eq '%') || ($Param[3] eq 'forked')) {
			# Generate a unique filehandle to enable nesting
			my $HandleName = "HANDLE$Param[2]";
			open ($HandleName,"<$Param[4]") or shout ('critical', "Unable to open data file $Param[4]: $! .");
		#	select ($HandleName); $| = 1; select (STDOUT);
			shout ('verbose', "--- Opening file $Param[4]");
			$FileLine = 0;
			$TempPosition = 0;
			while (<$HandleName>) {
				$_ =~ s/[\r\n]//g;
				next if ($_ eq '');
				$TempPosition++;
				next if ($LastPosition > $TempPosition);
				if (($Param[3] eq 'forked') && ($VAR{instances} ne '')) {
					# Guys, we are forked, so we skip lines
					$Modulus = $FileLine % $VAR{instances};
					$FileLine++;
					next if (($Modulus + 1) != $VAR{instance});
				}
				$SUBST{$Param[2]} = $_;
				&ExecProc ($Param[1]);
				if ($VAR{autoresume} eq 'on') {
					$TEMPFILE = '>'.$VAR{tmpfile}.$VAR{instance};
					open TEMPFILE or die 'gooofed';
					print TEMPFILE $TempPosition;
					close TEMPFILE;
				}
			}
			shout ('debug', "CALL: File $Param[2] exhausted");
			close $HandleName;
			unlink $VAR{tmpfile}.$VAR{instance};
			return;
		} elsif ($Param[3] eq '1way') {
			# Get the data from 1way process
			my $FileHandle = "EXTERNAL$Param[2]";
			$DATAPROC = "$ParamAfter[4]|";
			open ($FileHandle, $DATAPROC) or shout ('critical', "Unable to spawn external $ParamAfter[4]: $! .");
			shout ('debig',"OPEN: One-way process spawned.");
			while (<$FileHandle>) {
				$_ =~ s/[\r\n]//g;
				next if ($_ eq '');
				$SUBST{$Param[2]} = $_;
				&ExecProc ($Param[1]);
			}
			close $FileHandle;
			return;
		}
	# End call processing, begin return processing
	} elsif ($Param[0] eq 'return') {
		$EndProc = 1;
		shout ('debug', "RETURN: Set EndProc to $EndProc");
		return;
	# End return processing, begin foreach processing
	} elsif ($Param[0] eq 'foreach') {
		# Push the data from the file specified
		$VarFile = $Param[2];
		$DATAFILE = "<$VarFile";
		open DATAFILE or shout ('critical', "Unable to open data file $VarFile: $! .");
		shout ('debug', "Getting data for subst $Param[1] from $VarFile.");
		while (<DATAFILE>) {
			$_ =~ s/[\r\n]//g;
			next if ($_ eq '');
			$SUBST{$Param[1]} = $_;
			shout ('debug', "$VarFile gave $_ for subst $Param[1].");
			&ProcessCommand ($ParamAfter[3]);
		}
		close DATAFILE;
		shout ('debug', "Closing $VarFile.");
		return;
	} elsif ($Param[0] eq 'exit') {
		print " Exiting ... \n";
		exit;
	} elsif ($Param[0] eq 'print') {
		if ($Param[1] eq 'cookies') {
			while (($RawCookieName,$RawCookieVal) = each %COOK) {
				shout ('raw', "Set-Cookie: $RawCookieName=$RawCookieVal; domain=$COOKD{$RawCookieName}; path=$COOKD{$RawCookieName}\n");
			}
		} else {
			shout ('user', $ParamAfter[1]);

		}
		return;
	} elsif ($Param[0] eq 'printfile') {
		open (OUTFILE, ">>$Param[1]") or shout ('critical', "FILE: Unable to open $Param[1]");
		print OUTFILE "$ParamAfter[2]\n";
		close OUTFILE;
	} elsif ($Param[0] eq 'printraw') {
		shout ('raw', $ParamAfter[1]);
		return;
	} elsif ($Param[0] eq 'continue') {
		return;
	} elsif ($Param[0] eq 'hostmap') {
		if ($Param[3] ne '-') {
			$HOSTMAP{"$Param[1]:$Param[2]"} = $Param[3];
			$PORTMAP{"$Param[1]:$Param[2]"} = $Param[4];
		} else {
			delete $HOSTMAP{"$Param[1]:$Param[2]"};
			delete $PORTMAP{"$Param[1]:$Param[2]"};
		}
		return;
	} elsif ($Param[0] eq 'noproxy') {
		if ($Param[3] ne '-') {
			$NOPROXY{"$Param[1]:$Param[2]"} = 'on';
		} else {
			delete $NOPROXY{"$Param[1]:$Param[2]"};
		}
		return;
	} elsif ($Param[0] eq 'stats') {
		# Printing statistics
		$lenclock = time() - $startclock;
		$lenclock = 1 if ($lenclock == 0);
		($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($lenclock);
		$density = sprintf ("%d", ($VAR{tot_req} / $lenclock));
		$TotalBytes = $VAR{tot_sent} + $VAR{tot_rcv};
		shout ('user', "TIME:            $hour hrs / $min min / $sec sec - total $lenclock seconds.");
 		shout ('user', "STAT: Requests:  $VAR{tot_req} total, $SuccessfulRequests successful - $density requests / sec.");
		shout ('user', "STAT: Responses: $VAR{tot_ok} OK, $RedirRequests redirects.");
		shout ('user', "STAT: Bytes:     $TotalBytes total, $VAR{tot_sent} sent.");
		$speed = sprintf ("%d", ($VAR{tot_rcv} / $lenclock));
		shout ('user', "STAT: Received:  $VAR{tot_rcv} bytes - $speed bytes / sec.");
		return;
	} else {
		shout ('critical', "Syntax error: $ParamAfter[0] .");
	}
}

sub ExecProc {
	shout ('debug', "PROC: Executing proc \' $_[0] \'.");
	for $Index (0..$#{$PROC{$_[0]}}) {
		$EndProc = 0;
		&ProcessCommand ($PROC{$_[0]}[$Index]);
		shout ('debug', "PROC: Status of EndProc : $EndProc");
		last if ($EndProc == 1);
	}
	$EndProc = 0;
	shout ('debug', "PROC: We are done with \' $_[0] \'.");
}


sub ParseURLParts {
# This subroutine will try to canonize an absolute or relative URL

	if ($VAR{prefix} eq '') {$VAR{prefix} = $DEF{http};}

	if ($VAR{port} eq '') {$VAR{port} = $DEF{port};}

	if ($VAR{url} =~ /^$DEF{file}(.*?)\//) {
		print "This is a file URL! - $'\n ";
		$VAR{prefix} = $DEF{file};
		$VAR{request} = "\/$'";
	} elsif ($VAR{url} =~ /^($DEF{http}|$DEF{https})(.*)/) {
		# The URL is absolute. Parsing...
	
		$RestPortion = $2;
		$VAR{prefix} = $1;

		if ($RestPortion =~ /\//) {
			$RestPortion =~ /^(.*?)\/(.*)/;
			$HostPortion = $1;
			$VAR{request} = '/'.$2;
		} else {
			$HostPortion = $RestPortion;
			$VAR{request} = '/';
		}

		# Set ports depending on protocol

		if ($VAR{prefix} eq $DEF{https}) {
			if ($VAR{honorhttps} ne 'no') {
				$VAR{ssl} = 'on';
			}
			$VAR{port} = $DEF{sslport};
		} else {
			$VAR{ssl} = 'no';
			$VAR{port} = $DEF{port};
		}

		# If URL contains port, use it instead
		if ($HostPortion =~ /:/) {
			($VAR{host}, $VAR{port}) = split (':',$HostPortion);
		} else {
			$VAR{host} = $HostPortion;
		}
	} else {
		#The URL is relative
		if (!($VAR{url} =~ /^\// ) ) {
			# The URL does not begin with '/'
			shout ('debug', "URL: \' $VAR{url} \' is relative. Have that in mind.");
			if ($VAR{'request'} =~ /\/$/) {
				# The last URL was a directory, so we append
				$VAR{'request'} = $VAR{'request'}.$VAR{'url'};
			} else {
				# The last URL was a file, so we find the dir
				$VAR{'request'} =~ /(.*)\/(.*?)/;
				$VAR{'request'} = $1.'/'.$VAR{'url'};
			}
		} else {
			# The URL begins with '/' => server root
			$VAR{'request'} = $VAR{'url'};
		}

		if ($VAR{host} eq '') {
			shout ('debug', "URL: Host not specified. Defaulting to \' $DEF{host} \'.");
			$VAR{host} = $DEF{host};
		}

		#Glue parts together
		$VAR{url} = $VAR{prefix}.$VAR{host};
		if ($VAR{port} ne $DEF{port}) {
			$VAR{url} .= ':'.$VAR{port};
		}
		$VAR{url}.= $VAR{'request'};
	}

}

sub ParseHeaderParts {
	$RawResponse[0] =~ /(.*?) (.*?) (.*)/;
	$VAR{'protoversion'} = $1;$VAR{'retcode'} = $2;$VAR{'rettext'} = $3;

	if (($VAR{in_attack} > 0) && ($VAR{'retcode'} != '401') && ($VAR{'retcode'} != '407') &&
		($VAR{'retcode'} != '400') &&  ($VAR{'retcode'} != '500')
		&& ($VAR{'dictionary'} ne '')) {
		shout ('normal', "\n--- Successful pair: $VAR{user} : $VAR{autopass} . Deactivating dictionary.");
		$VAR{password} = $VAR{autopass};
		$VAR{dictionary} = '';
		$VAR{in_attack} = 0;
	}

	if (($VAR{in_attack} > 0) && (($VAR{'retcode'} eq '401') || ($VAR{'retcode'} eq '407')) &&
		($VAR{'dictionary'} ne '')) {
		# Display dots if we do a dictionary attack
		if ($VAR{shout} ne 'silent') {print STDOUT '.';}
	} else {
		# Display the response otherwise
		shout ('normal', "<<< Response: $VAR{retcode} $VAR{rettext}");
	}
	
	foreach $LineNum (1..$#RawResponse) {

		($RawHeader, $RawHeaderValue) = split (': ',$RawResponse[$LineNum]);

		shout ('debug', "HTTP: Header $RawHeader = $RawHeaderValue .");
		$VAR{$RawHeader} = $RawHeaderValue;

		# Extract the cookies
		if ($RawHeader eq 'Set-Cookie') {
			@CookieParams = ();

			# Split the cookie
			@CookieParts = split(';',$RawHeaderValue);

			if ($CookieParts[0] eq '') {
				# If we are sent a cookie without parameters, handle that
				shout ('error', "Malformed cookie received: $RawHeaderValue .");
			} else {
				# Extract name and value from the first part
				# This is not done with split, because I have
				# seen cookies with '=' in them.
				$CookieParts[0] =~ /(.*?)=(.*)/;
				$CookieName = $1; $CookieVal = $2;
				$COOK{$CookieName} = $CookieVal;

				# Extract cookie attributes
				foreach $CookiePart (@CookieParts) {
					($CookieParamName, $CookieParamValue) = split('=',$CookiePart);
					$CookieParamName =~ s/^[ \t]*//;	#Strip whitespace
					$CookieParams{lc($CookieParamName)} = lc($CookieParamValue);
				}
		
				# Place cookie parts in the arrays
				$COOKP{$CookieName} = $CookieParams{path};
				$COOKD{$CookieName} = $CookieParams{domain};
						
				shout ('verbose', "<<< Cookie: $CookieName = $CookieVal");
			}
               	}
	}
}

sub ParseRawResponse {

	&PrintToRaw;

	# Act upon a redirect
	if (($VAR{retcode} eq '301') or ($VAR{retcode} eq '302')) {
		$RedirRequests++;
		shout ('normal', "<<< Redirect: $VAR{Location} .");
		if ($VAR{autoredir} eq 'on') {
			$VAR{url} = $VAR{Location};
			if ($VAR{method} eq 'POST') {$VAR{method} = "GET";}
			%FLD = ();
			$AnyFields = 0;
			&SendRequestToHost;
		} else {
			shout ('debug', "HTTP: The redirect will not be followed automatically. Do it yourself.");
		}
		return;
	} elsif (($VAR{'retcode'} eq '401') || ($VAR{'retcode'} eq '407')) {

		if ($VAR{retcode} eq '407') {
			# Proxy requiring authentication
			$VAR{auth_header} = $VAR{'Proxy-Authenticate'};
			if (!($VAR{auth_header} =~ /(.*) realm=\"(.*?)\"/)) {
				shout ('error', "Unable to parse $VAR{auth_header}.");
				return;
			} else {
				$VAR{'ses_auth'} = $1;
				$VAR{'ses_realm'} = $2;
				$VAR{'auth_target'} = "Proxy-Authorization";
			}
		} else {
			# Web server requiring authentication
			$VAR{auth_header} = $VAR{'WWW-Authenticate'};
			if (!($VAR{auth_header} =~ /(.*) realm=\"(.*?)\"/)) {
				shout ('error', "Unable to parse $VAR{auth_header}.");
				return;
			} else {
				$VAR{'ses_auth'} = $1;
				$VAR{'ses_realm'} = $2;
				$VAR{'auth_target'} = "Authorization";
			}
		}

		if ($VAR{'ses_auth'} eq 'Digest') {
			if (!($VAR{auth_header} =~ /(.*) nonce=\"(.*?)\"/)) {
				shout ('error', "Unable to parse $VAR{'WWW-Authenticate'} .");
				return;
			} else {
				$VAR{'ses_nonce'} = $2;
			}
			shout ('debug', "DIGEST: Server nonce: $VAR{'ses_nonce'}.");
		}

		if ($VAR{'realm'} ne $VAR{'ses_realm'}) {
			shout ('error', "No user/pass pairs for realm \"$2\".");
			return;
		}

		if (($VAR{'dictionary'} ne '') && ($VAR{password} eq '')) {
			while (<DICTFILE>) {
				$VAR{'autopass'} = $_;
				if ($VAR{'autopass'} eq '') {
					shout ('error', "Dictionary file exhausted. $VAR{'tot_pass'} passwords tried.");
					return;
				}
				$VAR{'autopass'} =~ s/[\r\n]//g;
				$VAR{'tot_pass'}++;
				shout ('debug', "HTTPAUTH: Dictionary gave \' $VAR{'autopass'} \' as password.");
				$VAR{in_attack} = 1;
				&SendRequestToHost;
				last if (($VAR{retcode} ne '401') && ($VAR{retcode} eq '407'));
			}
		} else {
			if ($VAR{password} eq ''){
				shout ('error', "No HTTP password defined.");
				return;
			}
			if ($VAR{autopass} eq $VAR{password}) {
				shout ('error', "Server did not accept HTTP password.");
				return;
			} else {
				$VAR{autopass} = $VAR{password};
			}
			&SendRequestToHost;
		}

		return;

	} elsif ($VAR{'retcode'} =~ /^20/) {
		$VAR{'tot_ok'}++;
	} else {
		shout ('debug', "HTTP: I think $VAR{'retcode'} means an error.");
		&ProcessError;
	}
}

sub FindTag {
# This engine will scan HTML for tags and return the info contained within
# number, raw, and text are artificial tag attributes

	if ($VAR{'container'} ne '') {
		$WorkingSpace = $VAR{'container'};
		shout ('debug', "FINDTAG: Restricted in a container of ".length($VAR{'container'})." bytes.");
	} else {
		$WorkingSpace = $VAR{'body'};
		shout ('debug', "FINDTAG: Working on an HTML body of ".length($VAR{'body'})." bytes.");
	}

	$FoundTag = '%%NONE%%';
	$TagNumber = 0;

	while ($WorkingSpace =~ s/$TagBegin{$TagCode}(.*?)$RightBrack//is) {
		$TagRest = $'; $TagNumber++;
		$TagBody = $1;
		%TAGA = ();
		$TAGA{'body'} = $TagBegin{$TagCode}.$1.$RightBrack;
		if ($TagEnd{$TagCode} ne '') {
			# Tag has a close tag -> find the HTML within
			if ($TagRest =~ s/(.*?)$TagEnd{$TagCode}//is) {
				$TagText = $1;
				$TAGA{'raw'} = $TagText;
				$TagText =~ s/$LeftBrack(.*?)$RightBrack//igs;
				$TagText =~ s/  / /ig;
				$TagText =~ s/&nbsp;/ /ig;
				$TAGA{'text'} = $TagText;
			}
		}

		$TAGA{'number'} = $TagNumber;

		# Collect all attributes of the tag.
		while ($TagBody =~ s|^(([a-zA-Z][a-zA-Z0-9\.\-_]*)\s*)||) {
			$TagAttrib = $2;
			$AttribValue = '';
			if ($TagBody =~ s|(^=\s*([\"\'])(.*?)\2\s*)||s) {
				$AttribValue = $3;
			} elsif ($TagBody =~ s|(^=\s*([^\"\'>\s][^>\s]*)\s*)||) {
				$AttribValue = $2;
			}
			$TAGA{lc($TagAttrib)} = $AttribValue;
		}

		if ($TagCode eq 'hidden') {
			if (lc($TAGA{'type'}) eq 'hidden') {
				$FoundTag = $TAGA{$TagVal{$TagCode}};
			}
		} elsif ($TagCode eq 'refresh') {
			# We have to find a refresh tag.
			# it is in fact a meta http-equiv tag.
			if (lc($TAGA{'http-equiv'}) eq 'refresh') {
				($Garbage, $FoundTag) = split (/; url=/i,$TAGA{$TagVal{$TagCode}});
			}
		} else {
			# We are looking for an ordinary tag.
			# See if we match.
			if ($TAGA{$TagMethod} =~ /$TagMatch/) {
				$FoundTag = $TAGA{$TagVal{$TagCode}};
			}
		}
		if ($TagEnumerate > 0) {
			# We have to find all occurences of this tag so
			# we do not stop parsing here.
			if (lc($TAGA{'type'}) eq 'hidden') {
				$FLD{$TAGA{'name'}} = $TAGA{'value'};
				shout ('verbose', "<<< Field: $TAGA{'name'} = $FLD{$TAGA{'name'}}");
				$AnyFields = 1;
			}
		} else {
			if ($FoundTag ne '%%NONE%%') {
				return;
			}
		}
	}
}

sub FindHTML {
# The purpose of this routine is to isolate portions of HTML located
# between one string and another. There are many things to be improved.
# My goal is to be able to extract clean numbers from HTML for MRTG ;-)
	if ($VAR{'container'} ne '') {
		$WorkingSpace = $VAR{'container'};
	} else {
		$WorkingSpace = $VAR{'body'};
	}

	$FoundHTML = '%%NONE%%';
	if ($WorkingSpace =~ s/$BeginHTML(.*?)$EndHTML//is) {
		$FoundHTML = $1;
	}
}

sub FindLine {
	my $FindLineNumber = shift;
	if ($VAR{'container'} ne '') {
		my @FindLines = split(/\n/,$VAR{container});
	} else {
		my @FindLines = split(/\n/,$VAR{body});
	}
	if (!(defined $FindLines[$FindLineNumber])) {
		shout ('error', "Line number $FindLineNumber does not exist.");
		return undef;
	} else {
		chomp $FindLineNumber;
		shout ('debug', "HTML: Line number $FindLineNumber located.");
		return $FindLines[$FindLineNumber];;
	}
}

sub SSLSpawnTunnel {
	my $TmpCmdLine;
	
	$TmpCmdLine = $DEF{'sslcmd'};
	$TmpCmdLine =~ s/SSLHOST/$SSLRemoteHost/g;
	$TmpCmdLine =~ s/SSLPORT/$SSLRemotePort/g;
	$TmpCmdLine =~ s/SSLLOCAL/$VAR{'ssllocal'}/g;

	if ($DEF{'sslconsole'} > 0 ) {
		$SSLProcessFlags = CREATE_NEW_CONSOLE;
	} else {
		$SSLProcessFlags = DETACHED_PROCESS;
	}

	#Spawn SSL Tunnel.
	shout ('verbose', "Spawning SSL Tunnel to $VAR{'host'}:$VAR{'port'}. Local port is $VAR{'ssllocal'}. ");
	if ($^O =~ /Win/) {
		# Win32::Process::Create and Win32::Sleep are hidden
		# behind shout() so that the interpreter does not complain
		# about undeclated functions under non-Win32 systems.
		shout ('debug', Win32::Process::Create($SSLTunnelPID, $DEF{'sslapp'}, $TmpCmdLine, 0, $SSLProcessFlags, "."));
		# Sleeping a little bit is necessary to give stunnel.exe
		# time to start listening to the port. Otherwise, we will
		# not be able to connect and an error will occur.
		# Sorry for this race condition ...
		shout ('verbose', " Spawned. Waiting for 5 seconds ...");
		sleep 5;
	} else {
		`$TmpCmdLine`;
	}

	$SSLTunnelFunctional = 1;
}

sub SSLKillTunnel {

	return if ($SSLTunnelFunctional != 1);
	
	shout ('verbose', "Killing SSL Tunnel to $VAR{'host'}:$VAR{'port'}. ");
	if ($^O =~ /win/i) {
		$SSLTunnelPID->Win32::Process::Kill('') || die "Unable to kill SSL Tunnel $DEF{'sslapp'}. Kill it yourself.";
	} else {
		`killall $DEF{'sslapp'}`;
	}

	$SSLTunnelFunctional = 0;
}

sub RequestFileURL {
	my $LocalProg = '';
	my $LocalBin = '';

	if ($VAR{request} =~ / /g) {
		$LocalBin = $`;
	} else {
		$LocalBin = $VAR{request};
	}

	if (!(-e $LocalBin)) {
		$RawResponse[0] = 'LOCAL/1.0 404 Local File Not Found';
		&ParseHeaderParts;
	} elsif (-B $LocalBin) {
		$RawResponse[0] = 'LOCAL/1.0 200 OK Local Binary Executed';
		$LocalProg = $VAR{request};
	} else {
		$VAR{filesize} = -s $VAR{request};
		print "SIZE: $VAR{filesize}\n";
		open (FILEURL, "<$VAR{request}");
		$VAR{shebang} = <FILEURL>;
		print "SHEBANG: $VAR{shebang}\n";
		if ($VAR{shebang} =~ /^#!(.*)/) {
			# We have a valid shebang
			$LocalProg = "$1 $VAR{request}";
			$RawResponse[0] = 'LOCAL/1.0 200 OK Local Script Executed';
		} else {
			$RawResponse[0] = 'LOCAL/1.0 200 OK Local File';
		}
	}

	if ($LocalProg ne '') {
		# We have script of some sort
		$ENV{REQUEST_URI} = $VAR{url};
		$ENV{REQUEST_METHOD} = $VAR{method};
		$ENV{HTTP_USER_AGENT} = $VAR{agent};
		$ENV{REMOTE_ADDR} = $VAR{remoteaddr};
#		$ENV{HTTP_REFERER} = $VAR{referer};
		if ($VAR{method} eq 'POST') {
			$ENV{CONTENT_LENGTH} = length $VAR{allfields};
		} else {
			$ENV{QUERY_STRING} = $VAR{allfields};
			$ENV{CONTENT_TYPE} = 'application/x-www-form-urlencoded';
		}
		$ExtPid = open2(*LocalRead, *LocalWrite, $LocalProg);
		select(LocalRead); $| = 1; select(STDOUT);
		if ($VAR{method} eq 'POST') {
			print LocalWrite $VAR{allfields};
		}
		while (<LocalRead>) {
			$VAR{body}.= $_;
		}
		close LocalRead;
		close LocalWrite;
	} else {
		# It appears that it is a plain old file
		open (FILEURL, $VAR{request});
		read (FILEURL, $VAR{body}, $VAR{filesize});
		close FILEURL;
	}
	&ParseHeaderParts;
}

sub SendRequestToHost {
# This is the routine that is responsible for sending requests to hosts.
# One of my problems here is to implement some timeouts.

	if ($VAR{in_attack} < 1) {
		&ParseURLParts;
	}

	if ($VAR{'beforerequest'} ne '') {&ProcessCommand ($VAR{'beforerequest'})};
	$VAR{allfields} = '';
	$VAR{'container'} = '';
	$VAR{'body'} = '';
	$VAR{'Location'} = '';
	$VAR{'Set-cookie'} = '';
	@RawResponse = ();
	$RawHeaders = '';
	$PrevRawHost = $RawHost;
	$PrevRawPort = $RawPort;
	$VAR{'tot_req'}++;

	if ($AnyFields > 0 ) {
		# Build QUERY string
		while (($key,$value) = each %FLD) {
			$VAR{allfields}.="$key=$value&";
			shout ('verbose', ">>> Field: $key = $value");
		}
		$AnyFields = 0;
		$VAR{allfields} =~ s/ /+/g;
		$VAR{allfields} =~ s/&$//g;
	}

	if (($VAR{'proxyhost'} ne '') && ($NOPROXY{"$VAR{'host'}:$VAR{'port'}"} eq '')) {
		# If using proxy, send entire URL
		$RawRequest = $VAR{'url'};
	} else {
		# If not, send just the relative URL
		$RawRequest = $VAR{'request'};
	}

	if ($BogusField > 0) {
		$VAR{allfields}.= 'F' x $BogusFieldSize;
		$VAR{allfields}.= '=';
		$VAR{allfields}.= 'V' x $BogusTextSize;
		$VAR{allfields}.= '&';
		$BogusField = 0;
	}

	if (($VAR{'method'} eq 'GET') && ($VAR{allfields} ne '')){

		if (!($RawRequest =~ /\?/)) {
			$RawRequest.='?';
		}

		# Append fields to the URL
		$RawRequest.=$VAR{allfields};
	}

	# Perform anti-IDS encoding (courtesy of r.f.p)
	if($VAR{'encode'} eq 'on') {
		$RawRequest =~ s/([-a-zA-Z0-9.])/sprintf("%%%x",ord($1))/ge;
	}

	# The first line of the request
	$RawRequest = "$VAR{'method'} $RawRequest";
	if (($VAR{in_attack} < 1) || ($VAR{'dictionary'} eq '')) {
		shout ('normal', ">>> Request : $RawRequest");
	}
	$RawRequest.= " $DEF{'protocol'}\n";

	# Add the Host: part of the request
	%RAWHEADERS = ();
	$RAWHEADERS{'Host'} = "$VAR{'host'}:$VAR{'port'}";

	# Add the user agent to the request
	if ($VAR{sendagent} eq 'on') {$RAWHEADERS{'User-Agent'} = $VAR{agent};}

	# Add Referer
	if (($DEF{sendreferer} ne 'no') && ($VAR{Referer} ne '')) {
		$RAWHEADERS{Referer} = $VAR{Referer};
	}

	# Add authentication
	if ($VAR{autopass} ne '') {
		if ($VAR{'ses_auth'} eq 'Digest') {
			my $md5 = new Digest::MD5;
			my(@digest);
			$md5->add(join(":", $VAR{user},$VAR{'realm'},$VAR{'autopass'}));
			push(@digest, $md5->hexdigest);
			$md5->reset;
			push(@digest, $VAR{'ses_nonce'});
			$md5->add(join(":",$VAR{'method'},$VAR{'request'}));
			push(@digest, $md5->hexdigest);
			$md5->reset;
			$md5->add(join(":", @digest));
			my($digest) = $md5->hexdigest;
			$md5->reset;
			$RAWHEADERS{$VAR{'auth_target'}} = "Digest username=\"$VAR{user}\", realm=\"$VAR{'realm'}\", nonce=\"$VAR{'ses_nonce'}\", uri=\"$VAR{'request'}\", response=\"$digest\"";
		} elsif ($VAR{'ses_auth'} eq 'Basic') {
			$AuthEncoded = encode_base64($VAR{user}.':'.$VAR{'autopass'});
			$AuthEncoded =~ s/[\r\n]//g;
			$RAWHEADERS{$VAR{'auth_target'}} = "Basic $AuthEncoded";
		}
	}

	# Add cookies destined for this path and domain
	$CookieLine = '';
	while (($RawCookieName,$RawCookieVal) = each %COOK) {
		if ($COOKD{$RawCookieName} eq '') {
			shout ('verbose', ">>> Our cookie: $RawCookieName = $RawCookieVal");
			$CookieLine.= "$RawCookieName=$RawCookieVal; ";
			next;
		}

		if ($VAR{'request'} =~ /^$COOKP{$RawCookieName}/) {
			if ($VAR{'host'} =~ /$COOKD{$RawCookieName}/) {
				shout ('verbose', ">>> Their cookie: $RawCookieName = $RawCookieVal");
				$CookieLine.= "$RawCookieName=$RawCookieVal; ";
			}
		}
	}

	if ($CookieLine ne '') {$RAWHEADERS{'Cookie'} = $CookieLine;}

	if ($VAR{'keepalive'} eq 'on') {
		$RAWHEADERS{'Connection'} = "Keep-Alive";
		$RAWHEADERS{'Proxy-Connection'} = "Keep-Alive";
	} else {
		$RAWHEADERS{'Connection'} = "close";
	}

	# Add user-defined headers
	while (($key,$value) = each %HEADER) {
		if ($value ne '') {
			$RAWHEADERS{$key} = $value;
		} else {
			delete $RAWHEADERS{$key};
		}
	}
	%HEADER = ();

	while (($key,$value) = each %RAWHEADERS) {
		$RawRequest.= "$key: $value\n";
	}

	# Add various bogus headers to the request
	if ($DEF{'headers'} ne '') {$RawRequest.= $DEF{'headers'};}
	
	# Add POST data
	if ($VAR{'method'} eq 'POST') {
		# Append fields to the request.
		$RawRequest.="Content-Type: application/x-www-form-urlencoded\n";
		$RawRequest.="Content-Length: ".length($VAR{allfields})."\n\n";
		$RawRequest.="$VAR{allfields}\n";
	}

	# Terminate request
	$RawRequest.="\n";

	# If this is a file:// URL, handle that separately
	if ($VAR{prefix} eq $DEF{file}) {
		&RequestFileURL;
		return;
	}

	# Perform host:port mapping, if any
	if ($HOSTMAP{"$VAR{'host'}:$VAR{'port'}"} ne '') {
		$AfterMapHost = $HOSTMAP{"$VAR{'host'}:$VAR{'port'}"};
		$AfterMapPort = $PORTMAP{"$VAR{'host'}:$VAR{'port'}"};
		shout ("debug", "SOCKET: $VAR{'host'}:$VAR{'port'} will be mapped to $AfterMapHost:$AfterMapPort.");
	} else {
		$AfterMapHost = $VAR{'host'};
		$AfterMapPort = $VAR{'port'};
	}

	# And finally, determine the actual host:port to connect to
	if ($VAR{'ssl'} eq 'on' ) {
		#If it is a SSL request, send through tunnel
		$RawHost = '127.0.0.1';
		$RawPort = $VAR{'ssllocal'};
		$SSLRemoteHost = $AfterMapHost;
		$SSLRemotePort = $AfterMapPort;
		&SSLSpawnTunnel;
	} elsif (($VAR{'proxyhost'} ne '') && ($NOPROXY{"$VAR{'host'}:$VAR{'port'}"} eq '')) {
		# If through proxy, send through proxy
		$RawHost = $VAR{'proxyhost'};
		$RawPort = $VAR{'proxyport'};
		shout ('debug', "SOCKET: Using HTTP proxy $RawHost:$RawPort");
	} else {
		$RawHost = $AfterMapHost;
		$RawPort = $AfterMapPort;
	}

	# Initialize counters and clean up variables
	$VAR{ses_rcv} = 0;	$VAR{ses_start} = Win32::GetTickCount;
	if ($VAR{ses_start} == 0) {$VAR{ses_start} = (time() * 1000);}

	$VAR{'Content-Length'} = '';		$VAR{'Location'} = '';

	# Only open the socket if it is necessary
	if ($VAR{'keepalive'} eq 'on') {
		if (($PrevRawHost ne $RawHost) || ($PrevRawPort ne $RawPort)) {
			# We are connecting to a new host, so we
			# will open the socket again
			shout ('debug', "SOCKET: Making new connection to $RawHost:$RawPort.");
			&CloseTheSocketForMe;
		} else {
			shout ('debug', "SOCKET: Using Keep-Alive connection to $RawHost:$RawPort.");
		}
		if ($SocketOpened != 1) {
			OpenTheSocketForMe ();
			if ($SocketOpened == 0) {return;}
		}
	} else {
		OpenTheSocketForMe ();
		if ($SocketOpened == 0) {return;}
	}

	# Send the request to the host.
	if (!(print $REMOTE $RawRequest)) {
		shout ('critical',"SOCKET: Unable to send data: $! .");
	} else {
		if ($VAR{shout} eq 'debug') {
			$DebugRequest = $RawRequest;
			$DebugRequest =~ s/\n/ <CR> /g;
			shout ('debug', "SOCKET: Sending: $DebugRequest");
		}
	}

	# Consume the header of the response
	while (<$REMOTE>) {
		$VAR{'ses_rcv'} = $VAR{'ses_rcv'} + length ($_);
		$_ =~ s/[\r\n]//g;
		last if ($_ eq '');
		push @RawResponse, $_;
	}
	&ParseHeaderParts;

	# If there is body, consume it, based on rcvmethod
	$VAR{body} = '';
	if (($VAR{method} ne 'HEAD') && ($VAR{rcvmethod} ne 'forcehead')) {
		if ($VAR{rcvmethod} eq 'binary') {
			if ($VAR{'Content-Length'} ne '') {
				read ($REMOTE, $VAR{body}, $VAR{'Content-Length'});
				shout ('debug', "HTTP: Accepted a response of $VAR{'Content-Length'} binary bytes.");
			} else {
				read ($REMOTE, $VAR{body}, $VAR{maxbytes});
				shout ('debug', "HTTP: Accepted a response of ".length($VAR{body})." binary bytes.");
			}
		} elsif ($VAR{rcvmethod} eq 'forcebytes') {
			shout ('debug', "HTTP: Will get only the first $VAR{maxbytes} bytes from response.");
			read ($REMOTE, $VAR{'body'}, $VAR{maxbytes});
			shout ('debug', "HTTP: Got the first $VAR{maxbytes} bytes from response.");
		} elsif ($VAR{rcvmethod} eq 'forcelines') {
			my $CurrentLine = 0;
			shout ('debug', "HTTP: Will get only the first $VAR{maxlines} from response.");
			while (<$REMOTE>) {
				$CurrentLine++;
				$VAR{'ses_rcv'} = $VAR{'ses_rcv'} + length ($_);
				$VAR{'body'} .= $_;
				last if $CurrentLine == $VAR{maxlines};
			}
		} elsif ($VAR{'Content-Length'} eq '') {
			while (<$REMOTE>) {
				$VAR{'ses_rcv'} = $VAR{'ses_rcv'} + length ($_);
				$VAR{'body'} .= $_;
			}
		} else {
			read ($REMOTE, $VAR{'body'}, $VAR{'Content-Length'});
		}
	}

	# Close the socket if necessary
	if (($VAR{'Connection'} eq 'close') || ($VAR{'keepalive'} ne 'on')
		|| ($VAR{'Connection'} eq '') || ($VAR{'rcvmethod'} eq 'forcehead')) {
		&CloseTheSocketForMe;
	}

	# Calculate various timings and numbers
	$VAR{ses_end} = Win32::GetTickCount;
	if ($VAR{ses_end} == 0) {$VAR{ses_end} = (time() * 1000);}
	$VAR{ses_time} = (($VAR{ses_start} - $VAR{ses_end}) / 1000);
	if ($VAR{'ses_time'} == 0) {$VAR{'ses_time'} = 1};
	$VAR{ses_sent} = length ($RawRequest);
	$VAR{tot_sent} = $VAR{tot_sent} + $VAR{ses_sent};
	$VAR{'tot_rcv'} = $VAR{'tot_rcv'}+ $VAR{'ses_rcv'};
	$VAR{'ses_speed'} = $VAR{'ses_rcv'} / $VAR{'ses_time'};
	$VAR{'ses_speed'} = sprintf ("%d", $VAR{'ses_speed'});
	$SuccessfulRequests++;

	&SSLKillTunnel;

	if ($RawResponse[0] ne '') {
		if ($VAR{in_attack} < 1) {
			$VAR{Referer} = $VAR{url};
			&ParseRawResponse;
		}
	} else {
		shout ('error', "Server did not return anything. Whose bug?");
	}
}

sub PrintToRaw {
	if ($VAR{'rawfile'} ne '') {
		open FILEOUT, ">>$VAR{rawfile}" or die "Unable to open raw file $VAR{'rawfile'}: $!";
		if ($VAR{rcvmethod} eq 'binary') {
			binmode FILEOUT;
			print FILEOUT $VAR{body};
			shout ('verbose', "--- Saved a binary of ".length($VAR{body}). " bytes. ");
		} else {
			if ($VAR{'dumprequest'} ne '') {
				shout ('debug', "RAWFILE: Dumping the HTTP request to $VAR{rawfile} .");
				print FILEOUT "$RawRequest";
			}
	
			if ($VAR{'dumpheaders'} ne '') {
				shout ('debug', "RAWFILE: Dumping the HTTP headers to $VAR{rawfile} .");
				foreach $RawLine (@RawResponse) {print FILEOUT "$RawLine\n";}
				print FILEOUT "\n";
			}

			if ($VAR{'dumpbody'} ne '') {
				shout ('debug', "RAWFILE: Dumping the HTTP body to $VAR{rawfile} .");
				print FILEOUT $VAR{'body'};
			}

			print FILEOUT "\n";
		}
		close FILEOUT;
	}
}

sub ProcessError {
	shout ('debug', "ERRORHANDLE: Will now do \' $VAR{'onerror'} \'.");
	if ($VAR{'onerror'} eq 'exit') {
		shout ('critical', "If you do not want to die on error, use \'var onerror = continue\'. "); 
	} elsif ($VAR{'onerror'} eq '') {
		shout ('critical', "Error handler not defined. At least use \' var onerror = continue \'.");
	} elsif ($VAR{'onerror'} ne 'continue') {
		&ProcessCommand($VAR{'onerror'});
	} 
}

sub shout {
	$NoiseLevel = shift;
	$ShoutLine = shift;
	if ($NoiseLevel eq 'debug'){
		return if ($VAR{shout} ne 'debug');
		print STDOUT "DEBUG: $ShoutLine\n";
	} elsif ($NoiseLevel eq 'critical') {
		print STDOUT "!!! $ShoutLine";
		print STDOUT " Terminating ...\n";
		exit;
	} elsif ($NoiseLevel eq 'user') {
		print STDOUT "=== $ShoutLine\n";
	} elsif ($NoiseLevel eq 'raw') {
		$ShoutLine =~ s/\\n/\n/g;
		$ShoutLine =~ s/\\t/\t/g;
		print STDOUT $ShoutLine;
	} elsif ($VAR{shout} eq 'silent') {
		return;
	} elsif ($NoiseLevel eq 'error') {
		print STDOUT "!!! $ShoutLine\n";
		&ProcessError;
	} elsif ($NoiseLevel eq 'info') {
		print STDOUT "--- $ShoutLine\n";
	} elsif (($NoiseLevel eq 'verbose') && ($VAR{shout} eq 'verbose')) {
		print STDOUT "$ShoutLine\n";
	} elsif ($NoiseLevel eq 'normal') {
		print STDOUT "$ShoutLine\n";
	}
}

sub OpenTheSocketForMe {
	$SocketOpened = 0;
	if ($VAR{'sockshost'} ne '') {
		my $tmpsock = new Net::SOCKS(
			socks_addr => $VAR{'sockshost'},
			socks_port => $VAR{'socksport'},
			protocol_version => $VAR{'socksversion'}
		);

		if ($VAR{'socksuser'} ne '') {
			$tmpsock->param('user_id',$VAR{'socksuser'});
			$tmpsock->param('user_password',$VAR{'sockspassword'});
			$tmpsock->param('force_nonanonymous', 1);
		}
		shout ('debug', "SOCKET: Connecting to $RawHost:$RawPort via SOCKS proxy at $VAR{'sockshost'}:$VAR{'socksport'}.");
		$SOCK = $tmpsock->connect(peer_addr => $RawHost, peer_port => $RawPort);
		if ($tmpsock->param('status_num') != SOCKS_OKAY) {
			$DaMessage = Net::SOCKS::status_message($tmpsock->param('status_num'));
			shout ('critical', "SOCKET: Problem with SOCKS server $VAR{'sockshost'}:$VAR{'socksport'} - $DaMessage.");
		} else {
			shout ('debug', "SOCKET: Connected to $RawHost:$RawPort via SOCKS proxy at $VAR{'sockshost'}:$VAR{'socksport'}.");
		}
		$REMOTE = $SOCK;
	} else {
		shout ('debug', "SOCKET: Will attempt to connect to $RawHost:$RawPort");
		$Protocol = getprotobyname('tcp')||0;
		if(! ($RawIP = inet_aton($RawHost) ) ){
			shout ('error', "Unable to resolve $RawHost .");
			return;
		}
	
		if (! (socket(SOCK,PF_INET,SOCK_STREAM,$Protocol) ) ){
			shout ('critical', "Unable to create socket.");
		}
	
		if (! (connect(SOCK, sockaddr_in($RawPort, $RawIP)))) {
			shout ('error', "$RawHost does not respond on port $RawPort.");
			&SSLKillTunnel;
			return;
		}

		shout ('debug', "SOCKET: Connected to $RawHost:$RawPort.");
		select(SOCK);	$|=1;	select(STDOUT);
		$REMOTE = SOCK;
	}
	$SocketOpened = 1;
}

sub CloseTheSocketForMe {
	if ($SocketOpened == 0) {return;}
	if ($VAR{'sockshost'} ne '') {
		shout ('debug',"SOCKET: Closing SOCKS connection.");
		$SOCK->close();
	} else {
		shout ('debug',"SOCKET: Closing socket.");
		close(SOCK);
	}
	$SocketOpened = 0;
}

sub MakeComparison {
# This routine serves to evaluate 'if' statements

	my $First = shift;
	my $Condition = shift;
	my $Second = shift;
	my $Result = 0;

	if (($First =~ /$Second/si) && ($Condition eq '==')) {
		$Result = 1;	      
	} elsif ((!($First =~ /$Second/si)) && ($Condition eq '!=')) {
		$Result = 1;
	} elsif (($First > $Second) && ($Condition eq '>')){
		$Result = 1;
	} elsif (($First < $Second) && ($Condition eq '<')){
		$Result = 1;
	} elsif (($First eq '') && ($Condition eq 'is') && ($Second eq 'empty')){
		$Result = 1;
	} elsif (($First ne '') && ($Condition eq 'not') && ($Second eq 'empty')){
		$Result = 1;
	} else {
		shout ('debug', "IF: $First $Condition $Second is not true.");
	}

	if ($Result == 1) {
		shout ('debug', "IF: $First $Condition $Second is true.");
	}
	return $Result;
}
